"""Utility functions and classes"""
import logging

import colorlog

import sys
import inspect
import warnings
import pandas as pd
from typing import Union, Callable, Optional, Mapping, Any, Dict, Tuple
from types import ModuleType, MethodType
from weakref import WeakSet
from functools import partial, wraps


def get_logger(level=logging.INFO):
    # 创建logger对象
    logger = logging.getLogger()
    logger.setLevel(level)
    # 创建控制台日志处理器
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    # 定义颜色输出格式
    color_formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(levelname)s: %(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )
    # 将颜色输出格式添加到控制台日志处理器
    console_handler.setFormatter(color_formatter)
    # 移除默认的handler
    for handler in logger.handlers:
        logger.removeHandler(handler)
    # 将控制台日志处理器添加到logger对象
    logger.addHandler(console_handler)
    return logger


LOGGER = get_logger()

def _getdoc(c_or_f: Union[Callable, type]) -> Optional[str]:
    if getattr(c_or_f, '__doc__', None) is None:
        return None
    doc = inspect.getdoc(c_or_f)
    if isinstance(c_or_f, type) and hasattr(c_or_f, '__init__'):
        sig = inspect.signature(c_or_f.__init__)
    else:
        sig = inspect.signature(c_or_f)

    def type_doc(name: str):
        param: inspect.Parameter = sig.parameters[name]
        cls = getattr(param.annotation, '__qualname__', repr(param.annotation))
        if param.default is not param.empty:
            return f'{cls}, optional (default: {param.default!r})'
        else:
            return cls

    return '\n'.join(
        f'{line} : {type_doc(line)}' if line.strip() in sig.parameters else line
        for line in doc.split('\n')
    )



def _one_of_ours(obj, root: str):
    return (
        hasattr(obj, "__name__")
        and not obj.__name__.split(".")[-1].startswith("_")
        and getattr(
            obj, '__module__', getattr(obj, '__qualname__', obj.__name__)
        ).startswith(root)
    )

def _descend_classes_and_funcs(mod: ModuleType, root: str, encountered=None):
    if encountered is None:
        encountered = WeakSet()
    for obj in vars(mod).values():
        if not _one_of_ours(obj, root):
            continue
        if callable(obj) and not isinstance(obj, MethodType):
            yield obj
            if isinstance(obj, type):
                for m in vars(obj).values():
                    if callable(m) and _one_of_ours(m, root):
                        yield m
        elif isinstance(obj, ModuleType) and obj not in encountered:
            encountered.add(obj)
            yield from _descend_classes_and_funcs(obj, root, encountered)



def annotate_doc_types(mod: ModuleType, root: str):
    for c_or_f in _descend_classes_and_funcs(mod, root):
        c_or_f.getdoc = partial(_getdoc, c_or_f)



# the actual API
import warnings

warnings.filterwarnings("ignore")

import matplotlib.pyplot as plt

plt.rcParams["pdf.fonttype"] = 42
plt.rcParams["svg.fonttype"] = "none"

from trackc import tl, pl, pa
from trackc.gs import make_spec, savefig, tenon

__all__ = ["tl", "pl", "pa", "make_spec", "tenon"]

"""
import sys
sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['tl', 'pl', 'pa']})
from ._utils import annotate_doc_types
annotate_doc_types(sys.modules[__name__], 'trackc')
del sys, annotate_doc_types
"""




from trackc.pl.bed import bed_track
from trackc.pl.bedgraph_matrix import bdgmat_track
from trackc.pl.bigwig import bw_track
from trackc.pl.gene import gene_track
from trackc.pl.links import links_track
from trackc.pl.mapc import getData2Map, mapC
from trackc.pl.mapc_markline import mapc_markline
from trackc.pl.scale import multi_scale_track, scale_track
from trackc.pl.vhighlight import vhighlight
from trackc.pl.Virtual4C import virtual4C
from trackc.pl.zoomin import zoomin



from typing import Sequence, Union

import bioframe as bf
from matplotlib.axes import Axes

from .zoomin import _region_pos


def vhighlight(
    axs: Union[Sequence[Axes], Axes, None] = None,
    regions: Union[Sequence[str], str, None] = None,
    light_regions: Union[Sequence[str], str, None] = None,
    colors: Union[Sequence[str], None] = "yellow",
    alpha: float = 0.3,
):
    """
    Plot vhighlight track, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    regions: `str` | `str list`
        The raw genome regions, some of these regions will be selected to zoom in.
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``),
        which means the reverse region
    light_regions: `str` | `str list`
        regions to be zoomin, The format is the same as `regions`.
        the regions of `light_regions` should be located in the `regions`
    colors: `str list`
        the colors of the light_regions
    alpha: `float`

    Example
    -------
    >>> regions = ['18:47400000-48280000', '18:75280000-74030000']
    >>> light_regions = ['18:47950000-48280000', '18:75280000-74850000']

    >>> ten = tc.tenon(figsize=(6,1))
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='bottom', height=1, hspace=0.1)

    >>> tc.pl.vhighlight(axs=[ten.axs(0), ten.axs(1)], colors=['y', 'b'], regions=regions, light_regions=light_regions)
    >>> tc.pl.multi_scale_track(ten.axs(2), regions=regions, scale_adjust='Mb', intervals=2)
    """

    if isinstance(regions, str):
        regions = [regions]
    if isinstance(light_regions, str):
        light_regions = [light_regions]

    row_GRs = _region_pos(regions)
    light_GRs = _region_pos(light_regions)
    len_sum = sum(row_GRs["len"])

    row_GRs = row_GRs[
        ["chrom", "fetch_start", "fetch_end", "isReverse", "len", "ps", "pe"]
    ]
    row_GRs.columns = ["chrom", "start", "end", "isReverse", "len", "ps", "pe"]

    light_GRs["light_width"] = light_GRs["len"] / len_sum
    light_GRs = light_GRs[["chrom", "fetch_start", "fetch_end", "light_width"]]
    light_GRs.columns = ["chrom", "start", "end", "light_width"]

    row_GRs = row_GRs.reset_index()
    light_GRs = light_GRs.reset_index()
    del light_GRs["index"]
    del row_GRs["index"]

    light_r = bf.overlap(light_GRs, row_GRs)
    light_r = light_r.query("chrom_==chrom_")

    light_r["light_s"] = 0
    light_r_cp = light_r.copy()

    for i, row in light_r_cp.iterrows():
        if row["isReverse_"] == True:
            light_r.loc[i, "light_s"] = (
                row["ps_"] + abs(row["end"] - row["end_"]) / len_sum
            )
        else:
            light_r.loc[i, "light_s"] = (
                row["ps_"] + abs(row["start"] - row["start_"]) / len_sum
            )

    if isinstance(colors, list) == False:
        colors = [colors]

    if len(colors) < light_r.shape[0]:
        repeat_times = (light_r.shape[0] + len(colors) - 1) // len(colors)
        colors = (colors * repeat_times)[: light_r.shape[0]]

    for ix, aix in enumerate(axs):
        # trans = transforms.blended_transform_factory(
        #    aix.transAxes, aix.transAxes)
        aix.bar(
            x=light_r["light_s"],
            height=1,
            width=light_r["light_width"],
            transform=aix.transAxes,
            align="edge",
            color=colors,
            # zorder=50,
            clip_on=False,
            alpha=alpha,
        )



import logging
from typing import List, Sequence, Union

import cooler
import numpy as np
import pandas as pd


class GenomeRegion:
    region = None
    chrom = None
    start = None
    end = None
    fetch_start = None
    fetch_end = None
    length = None
    isReverse = False

    def __init__(self, region: str):
        self.region = region
        tmp = region.split(":")
        self.chrom = tmp[0]
        if len(tmp) == 2:
            self.start = int(tmp[1].split("-")[0])
            self.end = int(tmp[1].split("-")[1])
            if self.start > self.end:
                self.isReverse = True
                self.fetch_start = self.end
                self.fetch_end = self.start
            else:
                self.fetch_start = self.start
                self.fetch_end = self.end
            self.length = abs(self.start - self.end)

    def fetchRegion(self):
        fetch_r = self.chrom
        if self.fetch_start != None:
            fetch_r = fetch_r + ":" + str(self.fetch_start) + "-" + str(self.fetch_end)
        return fetch_r

    def GenomeRegion2df(self):
        """
        region4coolFetch = self.chrom + ":" + str(self.start) + '-' + str(self.end)
        if self.start == None:
            region4coolFetch = self.chrom
        else:
            if self.start > self.end:
                region4coolFetch = self.chrom + ":" + str(self.end) + '-' + str(self.start)
        """
        region4coolFetch = self.fetchRegion()

        df = pd.DataFrame(
            {
                "chrom": [self.chrom],
                "start": [self.start],
                "end": [self.end],
                "isReverse": [self.isReverse],
                "fetch_start": [self.fetch_start],
                "fetch_end": [self.fetch_end],
                "region4coolFetch": [region4coolFetch],
            },
            index=[self.region],
        )

        return df


class RegionsCmat:
    def __init__(
        self, cmat: np.array, row_regions: pd.DataFrame, col_regions: pd.DataFrame
    ):
        self.cmat = cmat
        self.row_regions = row_regions
        self.col_regions = col_regions


# import extractContactRegions as subsetContactRegions


def extractContactRegions(
    clr: Union[cooler.Cooler, str],
    balance: bool = False,
    # divisive_weights = None,
    row_regions: Union[Sequence[str], str, None] = None,
    col_regions: Union[Sequence[str], str, None] = None,
) -> RegionsCmat:
    """
    Extract a set of regions matrix from the cool format Hi-C matrix.

    The extracted matrix will splice intra and inter region interaction according to
    the given order and direction of the regions.

    Parameters
    ----------
    clr: `cooler.Cooler` | `str`
        cool format Hi-C matrix (https://github.com/open2c/cooler) or cool file path

        `cooler.Cooler` e.g. GM12878=cooler.Cooler('./GM12878.chr18.mcool::/resolutions/50000')

        `str` e.g. 'GM12878.chr18.mcool::/resolutions/50000' or 'GM12878.chr18.cool'

    balance: `bool`
        The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    row_regions: `str` | `str list` | None.
        The subset matrix row genome regions.
        e.g. ``"chr6:1000000-2000000"``, e.g. ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (e.g. ``"chr6:2000000-1000000"``),
        which means you want to get the reverse region contact matrix
    col_regions: `str` | `str list` | None.
        The subset matrix col genome regions, default is ``None``, which means the sample region as ``row_regions``

    Returns:
    --------
        :class:`~trackc.tl.RegionsCmat`
            row_regions and col_regions contact matrix object

    Example
    -------
    >>> import trackc as tc
    >>> import cooler

    >>> mat1 = tc.tl.extractContactRegions(clr='GM12878.chr18.mcool::/resolutions/50000', row_regions="18:45000000-78077248")
    >>> GM12878 = cooler.Cooler('./GM12878.chr18.mcool::/resolutions/50000')
    >>> mat2 = tc.tl.extractContactRegions(clr=GM12878, row_regions=["18:61140000-63630000", "18:74030000-77560000"], col_regions="18:47340000-50370000")
    >>> print(mat2.cmap.shape)
    >>> print(mat2.row_regions)
    >>> print(mat2.col_regions)
    """
    # divisive_weights: bool, optional
    #    Force balancing weights to be interpreted as divisive (True) or
    #    multiplicative (False). Weights are always assumed to be
    #    multiplicative by default unless named KR, VC or SQRT_VC, in which
    #    case they are assumed to be divisive by default.

    # -------
    if isinstance(row_regions, list):
        row_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in row_regions]
        )
    else:
        row_GenomeRegions = GenomeRegion(row_regions).GenomeRegion2df()

    if col_regions == None:
        col_GenomeRegions = row_GenomeRegions.copy()
    else:
        if isinstance(col_regions, list):
            col_GenomeRegions = pd.concat(
                [GenomeRegion(i).GenomeRegion2df() for i in col_regions]
            )
        else:
            col_GenomeRegions = GenomeRegion(col_regions).GenomeRegion2df()

    # ------
    if isinstance(clr, str):
        clr = cooler.Cooler(clr)

    region_mat_dic = {}
    for _, row_row in row_GenomeRegions.iterrows():
        for _, col_row in col_GenomeRegions.iterrows():
            row_col_region_cmat = clr.matrix(balance=balance).fetch(
                row_row["region4coolFetch"], col_row["region4coolFetch"]
            )
            if row_row["isReverse"] == True:
                row_col_region_cmat = np.flip(row_col_region_cmat, 0)
            if col_row["isReverse"] == True:
                row_col_region_cmat = np.flip(row_col_region_cmat, 1)

            region_mat_dic[
                "{0}__{1}".format(
                    row_row["region4coolFetch"], col_row["region4coolFetch"]
                )
            ] = row_col_region_cmat

    vstack_list = [None] * row_GenomeRegions.shape[0]
    hstack_list = [None] * col_GenomeRegions.shape[0]

    for i, row_region in enumerate(row_GenomeRegions["region4coolFetch"].to_list()):
        for ii, col_region in enumerate(
            col_GenomeRegions["region4coolFetch"].to_list()
        ):
            hstack_list[ii] = region_mat_dic["{0}__{1}".format(row_region, col_region)]

        vstack_list[i] = np.hstack(tuple(hstack_list))

    cMat = np.vstack(tuple(vstack_list))

    ####
    row_GenomeRegions["cbins"] = 0
    col_GenomeRegions["cbins"] = 0

    row_index = row_GenomeRegions.index.to_list()
    col_index = col_GenomeRegions.index.to_list()

    # print(row_index)
    # print(col_index)

    for i in row_index:
        for ii in col_index:
            cmat_shape = region_mat_dic[
                "{0}__{1}".format(
                    row_GenomeRegions.loc[i, "region4coolFetch"],
                    col_GenomeRegions.loc[ii, "region4coolFetch"],
                )
            ].shape

            row_GenomeRegions.loc[i, "cbins"] = cmat_shape[0]
            col_GenomeRegions.loc[ii, "cbins"] = cmat_shape[1]

    # print(col_GenomeRegions)
    row_GenomeRegions["chrom"] = row_GenomeRegions["chrom"].astype(str)
    col_GenomeRegions["chrom"] = col_GenomeRegions["chrom"].astype(str)
    # row_GenomeRegions['fetch_start'] = row_GenomeRegions['fetch_start'].astype(int)
    # col_GenomeRegions['fetch_start'] = col_GenomeRegions['fetch_start'].astype(int)

    r_l_regions_cMat = RegionsCmat(
        cmat=cMat, row_regions=row_GenomeRegions, col_regions=col_GenomeRegions
    )

    return r_l_regions_cMat


# import extractCisRegion as subsetCisRegion


def extractCisContact(
    clr: Union[cooler.Cooler, str],
    region: str,
    extend: int = 0,
    balance: bool = False,
    # divisive_weights = None,
) -> np.array:
    """
    Extract cis contact matrix from the cool or mcool format Hi-C matrix.

    Parameters
    ----------
    clr: `cooler.Cooler` | `str`
        cool format Hi-C matrix (https://github.com/open2c/cooler) or cool file path

        `cooler.Cooler` e.g. GM12878=cooler.Cooler('./GM12878.chr18.mcool::/resolutions/50000')

        `str` e.g. 'GM12878.chr18.mcool::/resolutions/50000' or 'GM12878.chr18.cool'
    region: `str`
        The subset matrix row genome regions
        eg. ``"chr6:1000000-2000000"`` or ``chr6``

    extend: `int`
        contact map extend to start and end position
    balance: `bool`
        The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``

    Returns:
    --------
    contact matrix: np.array
        matrix start with top-left

    Example
    -------
    >>> import trackc as tc
    >>> import cooler
    >>> mat1 = tc.tl.extractCisContact(clr='GM12878.chr18.mcool::/resolutions/50000', region="18:45000000-78077248")
    >>> GM12878 = cooler.Cooler('./GM12878.chr18.mcool::/resolutions/50000')
    >>> mat2 = tc.tl.extractCisContact(clr=GM12878, region="18:45000000-78077248")
    >>> print(mat2.shape)
    """

    if isinstance(clr, str):
        clr = cooler.Cooler(clr)

    resolution = clr.binsize
    genome_region = GenomeRegion(region)

    if genome_region.chrom not in clr.chromsizes:
        print(genome_region.chrom, " is not a chrom in the cool matrix")
        if genome_region.chrom.startswith('chr'):
            genome_region.chrom = genome_region.chrom.lstrip('chr')
        else:
            genome_region.chrom = 'chr' + genome_region.chrom
        print (f'{genome_region.chrom} instead')

    maxChromL = clr.chromsizes[genome_region.chrom]

    # ------
    if genome_region.fetch_start != None:
        genome_region.fetch_start = genome_region.fetch_start - extend * resolution
        if genome_region.fetch_start < 0:
            # genome_region.fetch_start = 0
            logging.error("Error: Extend start is less than 0, set extend=0 is ok\n")
            # sys.exit(1)

        genome_region.fetch_end = genome_region.fetch_end + extend * resolution
        if genome_region.fetch_end > maxChromL:
            # genome_region.fetch_end = maxChromL
            logging.error(
                "Error: extend is larger than chrom length, set extend=0 is ok\n"
            )
            # sys.exit(1)

    df = clr.matrix(balance=balance).fetch(genome_region.fetchRegion())
    return df



#!/Users/yuanzan/anaconda3/bin/python3
# -*- coding: utf-8 -*-

import click
import yaml

import warnings
warnings.filterwarnings("ignore")

import trackc as tc

from trackc.scripts.gtf2bed12 import gtf2bed
from trackc.scripts.trackc_cli import cli as track
#from trackc_cli import track
#from trackc.scripts.tadscore import tadScore


@click.group()
def main():
    pass

@main.command(name="cli", short_help="plot trackc by use yaml format configfile, yaml track type parameters is the same as API")
@click.argument('config', metavar='<trackc-conf.yml>')
@click.option('--regions', '-r', default=None,
              help="genome regions, \
              eg.:\'18:47950000-48280000 18:75280000-74850000\'")

@click.option('--outfile', '-o', default='trackc.pdf',
              help='output filename. default=trackc.pdf')

@click.option('--basefigsize', '-s', default='6,1',
              help='base figsize: width,height. default=\"6,1\"\
              The height option in the config.yml is relative to the base figsize height')

def cli(config, regions, outfile, basefigsize):
    track(config, regions, outfile, basefigsize)



@main.command(name="gtf2bed", short_help="Converts the gene GTF to bed12 or bed13 format") 
@click.argument('gtf', metavar='<genome.gtf>')
#, help="input file. GTF format gene annotation."
@click.option('--outfile', '-o', default='gene.bed',
              help='File name of gene annotation bed12 format.')
@click.option('--sep', '-s', default=';',
              help='tags of GTF sepration by which str')
@click.option('--gene_name_tag', '-nt', default='gene_name',
              help='tags of GTF gene_name')
@click.option('--gene_id_tag', '-it', default='gene_id',
              help='tags of GTF gene_id')
@click.option('--gene_biotype_tag', '-bt', default='gene_biotype',
              help='tags of GTF gene_biotype')
@click.option('--biotype2bed13', '-bed13', is_flag=True, show_default=True, default=True,
              help='if this arg is set, the out file columns 13 will add gene\'s biotype')

def mode2(gtf, outfile, sep, gene_name_tag, gene_id_tag, gene_biotype_tag, biotype2bed13):
	gtf2bed(gtf, outfile, sep, gene_name_tag, gene_id_tag, gene_biotype_tag, biotype2bed13)

#@main.command()
#@click.argument('mcool', metavar='name.mcool::/resolutions/10000')

#def tadscore(mcool):
#    tadScore()


if __name__ == '__main__':
    main()




from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import sys
import pandas as pd
from matplotlib.axes import Axes
import pyBigWig
from trackc._utils import LOGGER
from trackc.tl._getRegionsCmat import GenomeRegion
import warnings
warnings.filterwarnings("ignore")

def _make_multi_region_ax(ax, lineGenomeRegions):
    lineGenomeRegions["len"] = (
        lineGenomeRegions["fetch_end"] - lineGenomeRegions["fetch_start"]
    )
    lineGenomeRegions["ax_ratio"] = (
        lineGenomeRegions["len"] / lineGenomeRegions["len"].sum()
    )
    lineGenomeRegions["ax_x"] = (
        lineGenomeRegions["ax_ratio"].cumsum(axis=0) - lineGenomeRegions["ax_ratio"]
    )
    axs = [
        ax.inset_axes([row["ax_x"], 0, row["ax_ratio"], 1])
        for i, row in lineGenomeRegions.iterrows()
    ]
    for axi in axs:
        axi.axis("off")

    return axs


def bw_track(
    bw,
    ax: Optional[Axes] = None,
    regions: Union[Sequence[str], str, None] = None,
    binsize: Optional[int] = 50000,
    style: Optional[str] = "bar",
    summary_type: Union[str, None] = "mean",
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    primary_col: Union[Sequence[str], None] = "#3271B2",
    secondary_col: Union[Sequence[str], None] = "#FBD23C",
    alpha: Optional[float] = 1.0,
    invert_y: Optional[bool] = False,
    label: Optional[str] = None,
    label_rotation: Union[int, None] = 0,
    label_fontsize: Optional[int] = 9,
    tick_fontsize: Optional[int] = 7,
    tick_fl: Optional[str] = "%0.2f",
    ax_on: bool = False,
):
    """
    Plot bigwig signal track, support for multiple or reverse genome regions.

    Parameters
    ----------
    bw: `pyBigWig.open` query object, or bigwig file path
    ax: :class:`matplotlib.axes.Axes` object
    regions: `str` | `str list`
        The genome regions to show the signal.
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``),
        which means the reverse region
    binsize: `int`
        binsize divided to computing signal summary statistics
    type: `str`
        plot type, default='bar', options=['bar', 'line']
    summary_type: `str`
        Summary type (mean, min, max, coverage, std), default 'mean'.
    vmin: `float`
        the minimum range of values used to define the ylim
    vmax: `float`
        the maximum range of values used to define the ylim
    primary_col: `str`
        the signal bar color
    secondary_col: `str`
        the signal bar color for negative values
    alpha: `float`
        alpha of plot color
    invert_y: `bool`
        whether reverse the y-axis
    label: `str`
        the title of the track, will show on the left
    label_rotation: `int`
        the label text rotation
    label_fontsize: `int`
        the label text fontsize
    tick_fontsize: `int`
        values range ticks text fontsize
    tick_fl: `str`
        values range ticks retains a few decimal places
    ax_on: `bool`
        whether show the spines

    Example
    -------
    >>> import trackc as tc
    >>> import pyBigWig
    >>> H3K27ac = pyBigWig.open('./GSM4604189.bigwig')

    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='bottom', height=1, hspace=0.2)

    >>> regions = ['chr8:127000000-129200000', 'chr14:96500000-99300000']
    >>> tc.pl.bw_track(H3K27ac, ten.axs(0), regions=regions, vmax=20, label='H3K27ac', binsize=10000, color='tab:blue')
    >>> tc.pl.bw_track(H3K27ac, ten.axs(1), regions=regions, vmax=5, label='H3K27ac', binsize=10000, invert_y=True, ax_on=True)
    >>> tc.savefig('trackc_bigwig_track.pdf')
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    if isinstance(bw, str) == True:
        bw = pyBigWig.open(bw)

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(primary_col, list) == False:
        primary_col = [primary_col]
    if len(primary_col) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(primary_col) - 1) // len(primary_col)
        primary_col = (primary_col * repeat_times)[: line_GenomeRegions.shape[0]]

    if isinstance(secondary_col, list) == False:
        secondary_col = [secondary_col]
    if len(secondary_col) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(secondary_col) - 1) // len(secondary_col)
        secondary_col = (secondary_col * repeat_times)[: line_GenomeRegions.shape[0]]

    min_y = 0
    max_y = 0
    plot_bottom_line = True

    for i, row in line_GenomeRegions.iterrows():
        bins = int(row["len"] / binsize)
        if row["chrom"] not in bw.chroms():
            raw_chr = row["chrom"]
            if row["chrom"].startswith('chr'):
                row["chrom"] = row["chrom"].lstrip('chr')
            else:
                row["chrom"] = 'chr' + row["chrom"]
            if row["chrom"] not in bw.chroms():
                print(f'{raw_chr} not in bigwig chroms!')
                return
        plot_list = bw.stats(row["chrom"], int(row["fetch_start"]), int(row["fetch_end"]),  type=summary_type, nBins=bins)
        plot_list = [0 if v is None else v for v in plot_list]
        if style == "line":
            axs[i].plot(range(0, bins), plot_list, color=primary_col[i], alpha=alpha)
        else:
            plotvalues = pd.DataFrame({'v':plot_list}, index=range(0, bins))
            pos_value = plotvalues.query('v>=0')
            neg_value = plotvalues.query('v<0')
            if neg_value.shape[0]>0:
                plot_bottom_line = False 

            axs[i].bar(
                x=pos_value.index,
                height=pos_value['v'],
                width=1,
                #bottom=[0] * (pos_value.shape[0]),
                bottom=0,
                color=primary_col[i],
                align="edge",
                edgecolor=None,
                alpha=alpha,
            )
            
            axs[i].bar(
                x=neg_value.index,
                height=neg_value['v'],
                width=1,
                #bottom=[0] * (neg_value.shape[0]),
                bottom=0,
                color=secondary_col[i],
                align="edge",
                edgecolor=None,
                alpha=alpha,
            )

        right, left = bins, 0
        if row["isReverse"] == True:
            left, right = bins, 0
        axs[i].set_xlim(left, right)

        if min_y > min(plot_list):
            min_y = min(plot_list)

        if max_y < max(plot_list):
            max_y = max(plot_list)

    if vmin == None:
        vmin = min_y
    if vmax == None:
        vmax = max_y

    for axi in axs:
        if invert_y:
            axi.set_ylim(vmax, vmin)
        else:
            axi.set_ylim(vmin, vmax)
    va = "top"
    if invert_y:
        va = "bottom"
        ax.set_ylim(vmax, vmin)
    else:
        ax.set_ylim(vmin, vmax)

    if plot_bottom_line:
        ax.text(
            0,
            vmax,
            " [{0}, {1}]".format(tick_fl % vmin, tick_fl % vmax),
            verticalalignment=va,
            fontsize=tick_fontsize,
        )
        ax.set_yticks([])
        ax.set_yticklabels("")

    else:
        ax.set_yticks([vmin, 0, vmax])
        ax.set_yticklabels([f'{tick_fl % vmin}', '0', f'{tick_fl % vmax}'], fontsize=tick_fontsize)
        if invert_y:
            ax.set_yticks([vmax, 0, vmin])
            ax.set_yticklabels([f'{tick_fl % vmax}', '0', f'{tick_fl % vmin}'], fontsize=tick_fontsize)
            
    ax.set_ylabel(
        label,
        fontsize=label_fontsize,
        rotation=label_rotation,
        horizontalalignment="right",
        verticalalignment="center",
    )

    if ax_on == False:
        spines = ["top", "bottom", "left", "right"]
        if invert_y == True:
            if plot_bottom_line==True:
                del spines[0]
            else:
                del spines[2]
        else:
            if plot_bottom_line==True:
                del spines[1]
            else:
                del spines[2]
        for i in spines:
            ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels("")
    





from typing import Optional, Sequence, Union

import pandas as pd
from matplotlib.axes import Axes

from trackc.tl._getRegionsCmat import GenomeRegion

from .bigwig import _make_multi_region_ax


def gene_track(
    ax: Optional[Axes] = None,
    bed12: Union[pd.DataFrame, str] = None,
    regions: Union[Sequence[str], str, None] = None,
    track_type: Union[str, None] = "gene",
    show_label: Union[bool, Sequence[str], str] = True,
    pos_strand_gene_color: Union[str, None] = "#3366CC",
    neg_strand_gene_color: Union[str, None] = "#EECFA1",
    line: Union[int, None] = 1,
    gene_fontszie: Union[int, None] = 7,
    label: Optional[str] = None,
    label_rotation: Union[int, None] = 0,
    label_fontsize: Optional[int] = 12,
    ax_on: bool = False,
):
    """
    Plot gene track, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    bed12: `pd.DataFrame` | `str`
        gene annotation bed12 format `DataFrame` or `filepath`
        Bed12 files can be converted from GTF using `gtf2bed4trackc`.
        https://trackc.readthedocs.io/en/latest/tutorials/fileformats/bed12.html#gtf2bed4trackc
    regions: `str` | `str list`
        genome regions, format: `chrom:start-end`.
        e.g. ``['chr18:47950000-48280000', 'chr18:75280000-74850000']`` or ``"chr18:45000000-78077248"``.
        If the start is bigger than end, the genome region will be reversed
    track_type: `str`
        you can select one of the options: `gene` or `dendity`
        gene: gene track style
        dendity: gene density style. Under development
    show_label: `bool` | `str` | `str list`
        If the value is `False`, the gene name will not show
        If want show one gene, and hide others, just set the gene or gene list as the value, eg: `PIBF1` | `['PIBF1', 'KLF5']`
    pos_strand_gene_color: `str`
        positive strand gene name color
    neg_strand_gene_color: `str`
        negative strand gene name color
    line: `int`
        rows occupied by the genes in the region plotted
    gene_fontszie: `int`
        gene label fontszie
    label: `str`
        the title of the track, will show on the left
    label_rotation: `int`
        the label text rotation
    label_fontsize: `int`
        the label text fontsize
    ax_on: `bool`
        If True, top, left and right spines will show

    Returns
    -------
    None

    Example
    -------
    >>> import trackc as tc
    >>> regions = ['chr18:47950000-48280000', 'chr18:75280000-74850000']
    >>> gene_bed12 = '/path/GRCh38.84.bed12'

    >>> fig, axs = tc.make_spec(figsize=(7,2), height_ratios=[1])
    >>> tc.pl.gene_track(gene_bed12, ax=axs[0], regions=regions, line=12)
    >>> tc.savefig('trackc_gene_track.pdf')
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    ax.set_ylabel(
        label,
        fontsize=label_fontsize,
        rotation=label_rotation,
        horizontalalignment="right",
        verticalalignment="center",
    )

    ax.set_xticks([])
    ax.set_xticklabels("")
    ax.set_yticks([])
    ax.set_yticklabels("")
    if ax_on == False:
        spines = ["top", "bottom", "left", "right"]
        for i in spines:
            ax.spines[i].set_color("none")

    if isinstance(bed12, str):
        bed12 = pd.read_table(bed12, sep="\t", header=None)

    bed12 = bed12.iloc[:, 0:12]
    bed12.columns = [
        "chrom",
        "start",
        "end",
        "name",
        "score",
        "strand",
        "thickStart",
        "thickEnd",
        "itemRgb",
        "blockCount",
        "blockSizes",
        "blockStarts",
    ]
    bed12["blockSizes"] = bed12["blockSizes"].str.rstrip(",")
    bed12["blockStarts"] = bed12["blockStarts"].str.rstrip(",")
    bed12["chrom"] = bed12["chrom"].astype(str)
    chrom_names = bed12["chrom"].unique()
    
    for ix, row in line_GenomeRegions.iterrows():
        if row["chrom"] not in chrom_names:
            raw_chr = row["chrom"]
            if row["chrom"].startswith('chr'):
                row["chrom"] = row["chrom"].lstrip('chr')
            else:
                row["chrom"] = 'chr' + row["chrom"]
            if row["chrom"] not in chrom_names:
                print(f'{raw_chr} not in bigwig chroms!')
                return

        if track_type == "gene":
            _plot_gene(
                axs[ix],
                bed12,
                row["chrom"],
                row["fetch_start"],
                row["fetch_end"],
                needReverse=row["isReverse"],
                show_label=show_label,
                pos_strand_gene_color=pos_strand_gene_color,
                neg_strand_gene_color=neg_strand_gene_color,
                line=line,
                fontszie=gene_fontszie,
                ax_on=ax_on,
            )
        if track_type == "density":
            print("This gene type is developping")
        else:
            pass


def _plot_gene(
    ax,
    gene_bed,
    chrom,
    start,
    end,
    needReverse=False,
    show_label=True,
    pos_strand_gene_color="#3366CC",
    neg_strand_gene_color="#EECFA1",
    line=1,
    fontszie=5,
    ax_on=False,
):
    gene_bed = gene_bed[gene_bed["chrom"] == chrom]
    gene_bed_plot = gene_bed[
        ((gene_bed["start"] >= start) & (gene_bed["start"] <= end))
        | ((gene_bed["end"] >= start) & (gene_bed["end"] <= end))
    ]
    gene_bed_plot = gene_bed_plot.sort_values(by="end")
    # print(gene_bed_plot

    plot_gene_num = gene_bed_plot.shape[0]

    ii = 0
    head_length = (abs(end - start) / (line + 2)) / 5
    if line <= 3:
        head_length = (abs(end - start) / (line * 3)) / 10

    for i, row in gene_bed_plot.iterrows():
        # col = pos_strand_gene_color
        text_col = pos_strand_gene_color

        if row["strand"] == "-":
            # col = neg_strand_gene_color
            text_col = neg_strand_gene_color

        # text_col = col
        plot_y = ii % line

        ax.plot(
            (row["start"], row["end"]),
            (plot_y + 0.5, plot_y + 0.5),
            color="k",
            linewidth=1,
            solid_capstyle="butt",
        )
        starts = [int(x) for x in row["blockStarts"].split(",")]
        widths = [int(x) for x in row["blockSizes"].split(",")]

        ax.bar(
            x=starts,
            height=0.4,
            width=widths,
            bottom=plot_y + 0.3,
            edgecolor="k",
            linewidth=1,
            align="edge",
            color="k",
        )

        if row["start"] < start:
            row["start"] = start
        if row["end"] > end:
            row["end"] = end

        arrow_s = row["end"]
        dx = 0.3
        if row["strand"] == "-":
            arrow_s = row["start"]
            dx = -0.1
        ax.arrow(
            arrow_s,
            plot_y + 0.5,
            dx,
            0,
            overhang=0.5,
            width=0,
            head_width=0.28,
            head_length=head_length,
            length_includes_head=False,
            color=text_col,
            linewidth=0.5,
        )
        if isinstance(show_label, bool):
            if show_label == False:
                ii += 1
                continue
        if isinstance(show_label, str):
            if row["name"] != show_label:
                ii += 1
                continue
        if isinstance(show_label, list):
            if row["name"] not in show_label:
                ii += 1
                continue

        if (row["name"] in gene_bed_plot.iloc[-4:, :]["name"]) or (
            int(line / (ii + 1)) < 2
        ):
            ha = "right"
            genename = row["name"] + "  "
            xpos = row["start"]
            if needReverse:
                ha = "left"
                genename = "  " + row["name"]
                # xpos = row['end']
            ypos = plot_y + 0.5
            if line == 1:
                xpos = row["start"] + abs(row["start"] - row["end"]) / 2
                ypos = 0.8
                ha = "center"
            ax.text(
                xpos,
                ypos,
                genename + "  ",
                ha=ha,
                va="center",
                color=text_col,
                fontsize=fontszie,
            )
        else:
            ha = "left"
            genename = "  " + row["name"]
            xpos = row["end"]
            if needReverse:
                ha = "right"
                genename = row["name"] + "  "
                # xpos = row['start']

            ypos = plot_y + 0.5
            if line == 1:
                xpos = row["start"] + abs(row["start"] - row["end"]) / 2
                ypos = 0.8
                ha = "center"
            ax.text(
                xpos,
                ypos,
                genename,
                ha=ha,
                va="center",
                color=text_col,
                fontsize=fontszie,
            )

        ii += 1

    xlim_s = start
    xlim_e = end
    if needReverse == True:
        xlim_s = end
        xlim_e = start

    ax.set_xlim(xlim_s, xlim_e)
    ax.set_ylim(top=0, bottom=line)
    if plot_gene_num < line:
        ax.spines["bottom"].set_position(("data", plot_gene_num))

    if ax_on == False:
        spines = ["top", "bottom", "left", "right"]
        for i in spines:
            ax.spines[i].set_visible(False)
            # for i in ['left','top','right']:
            ax.spines[i].set_color("none")
            ax.spines[i].set_linewidth(0)
    ax.spines["bottom"].set_color("black")
    ax.spines["bottom"].set_linewidth(0.5)
    ax.tick_params(bottom=True, top=False, left=False, right=False)
    ax.set_xticklabels("")
    ax.set_yticklabels("")



from typing import Optional, Sequence, Union

import matplotlib.pyplot as plt
import numpy as np
import sys
import pandas as pd
from matplotlib import cm
from matplotlib.axes import Axes
from matplotlib.collections import PatchCollection
from matplotlib.colors import Colormap
from matplotlib.patches import Polygon

from trackc.pl.bigwig import _make_multi_region_ax
from trackc.pl.links import _plot_loop_arc
from trackc.tl._getRegionsCmat import GenomeRegion
from matplotlib.colors import TwoSlopeNorm

def bed_track(
    ax: Optional[Axes] = None,
    bed: Union[pd.DataFrame, str, None] = None,
    regions: Union[Sequence[str], str, None] = None,
    style: Union[str, None] = "bar",
    primary_col: Union[Sequence[str], None] = "#3271B2",
    secondary_col: Union[Sequence[str], None] = "#FBD23C",
    cmap: Union[Colormap, str, None] = None,
    intervals: Union[int, None] = 1,
    # show_names: Union[bool, None] = False,
    alpha: Union[float, None] = 1,
    invert_y: Optional[bool] = False,
    label: Union[str, None] = None,
    label_fontsize: Union[int, None] = 12,
    label_rotation: Union[int, None] = 0,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    tick_fontsize: Optional[int] = 8,
    tick_fl: Optional[str] = "%0.2f",
    score_label_size: Union[int, None] = 7,
):
    """
    Plot bed track, support for multiple or reverse genome regions.
    support bed3 and bed5, the fields after the column5 will be ignored,
    should be sorted py chromStart if ``style`` is `line`

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    bed: `pd.DataFrame` | `str`
        If ``bed`` if a filepath, the file should have no headers.

        Here is bed formats:
            column1: chrom
                The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold (e.g. scaffold10671).
            column2: chromStart
                The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 0.
            column3: chromEnd
                The ending position of the feature in the chromosome or scaffold.
            column4: name
                Defines the name of the BED line. Either "." (=no name), or other string
            column5: score
                Defines the name of the BED line. if nessasary, can be set as ``.``,
                if track_type/style is one of bar/line
            column6: strand
                Defines the strand. Either "." (=no strand) or "+" or "-".

            if the input data have 4 columns, then default input is bedGraph format: [chrom start end score]

    regions: `str` | `str list`
        The genome regions to plot.
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000"]``

        The start can be larger than the end (e.g. ``"chr6:2000000-1000000"``),
        which means you want to get the reverse region
    style: `str`
        bed blocks style,  opions in ['line', 'bar', 'link', 'tri', 'rec']
    primary_col: `str`
        the color of line/tri/rec, if color is color list, the block color will set by regions
    secondary_col: `str`
        the color of line/tri/rec for negative values
    cmap: `str` | `matplotlib.colors.Colormap`
        the colormap of the plot except style:line
    intervals: `int`
        if style is one of [tri, rec], the row number distribution for triangle or rectangle blocks
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()
    if isinstance(primary_col, list) == False:
        primary_col = [primary_col]
    if len(primary_col) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(primary_col) - 1) // len(primary_col)
        primary_col = (primary_col * repeat_times)[: line_GenomeRegions.shape[0]]
    if isinstance(secondary_col, list) == False:
        secondary_col = [secondary_col]
    if len(secondary_col) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(secondary_col) - 1) // len(secondary_col)
        secondary_col = (secondary_col * repeat_times)[: line_GenomeRegions.shape[0]]

    if isinstance(cmap, list) == False:
        cmap = [cmap]
    if len(cmap) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(cmap) - 1) // len(cmap)
        cmap = (cmap * repeat_times)[: line_GenomeRegions.shape[0]]

    ax.set_ylabel(
        label, fontsize=label_fontsize, rotation=label_rotation, ha="right", va="center"
    )
    spines = ["top", "right", "left", "bottom"]
    for i in spines:
        ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels("")
    ax.set_yticks([])
    ax.set_yticklabels("")

    if isinstance(bed, str) == True:
        bed = pd.read_table(bed, sep="\t", header=None)
    else:
        bed = bed.copy()

    bed = bed.fillna(0)
    score_label = ""
    if bed.shape[1] == 3:
        bed.columns = ["chrom", "start", "end"]
    if bed.shape[1] == 4:
        bed.columns = ["chrom", "start", "end", "score"]
        if bed["score"].dtypes == "int64":
            print("Input data is bedGraph format")
    if bed.shape[1] == 5:
        score_label = bed.columns[4]
        bed.columns = ["chrom", "start", "end", "name", "score"]
    if bed.shape[1] == 6:
        score_label = bed.columns[4]
        bed.columns = ["chrom", "start", "end", "name", "score", "strand"]
    bed["chrom"] = bed["chrom"].astype(str)

    min_y = None
    max_y = None
    max_len = 0

    chromos = bed['chrom'].unique()
    for ix, row in line_GenomeRegions.iterrows():
        raw_chr = row["chrom"]
        if row["chrom"] not in chromos:
            if row["chrom"].startswith('chr'):
                row["chrom"] = row["chrom"].lstrip('chr')
            else:
                row["chrom"] = 'chr' + row["chrom"]

            if row["chrom"] not in chromos:
                print(f'{raw_chr} not in beg chroms!')
                sys.exit(0)

        bed2plot = bed[
            (bed["chrom"] == row["chrom"])
            & (bed["end"] >= row["fetch_start"])
            & (bed["start"] <= row["fetch_end"])
        ]


        if style in ["line", "bar"] or bed.shape[1] >= 4:
            if min_y == None:
                min_y = bed2plot["score"].min(skipna=True, numeric_only=True)
            else:
                if min_y < bed2plot["score"].min(skipna=True, numeric_only=True):
                    min_y = bed2plot["score"].min(skipna=True, numeric_only=True)

            if max_y == None:
                max_y = bed2plot["score"].max(skipna=True, numeric_only=True)
            else:
                if max_y > bed2plot["score"].max(skipna=True, numeric_only=True):
                    max_y = bed2plot["score"].max(skipna=True, numeric_only=True)

        maxlength = (bed2plot["end"] - bed2plot["start"]).max(
            skipna=True, numeric_only=True
        )
        if max_len < maxlength:
            max_len = maxlength

    if vmin == None:
        vmin = min_y
    if vmax == None:
        vmax = max_y

    for ix, row in line_GenomeRegions.iterrows():
        raw_chr = row["chrom"]
        if row["chrom"] not in chromos:
            if row["chrom"].startswith('chr'):
                row["chrom"] = row["chrom"].lstrip('chr')
            else:
                row["chrom"] = 'chr' + row["chrom"]

            if row["chrom"] not in chromos:
                print(f'{raw_chr} not in beg chroms!')
                sys.exit(0)
                
        bed2plot = bed[
            (bed["chrom"] == row["chrom"])
            & (bed["end"] >= row["fetch_start"])
            & (bed["start"] <= row["fetch_end"])
        ].copy()
        if bed2plot.shape[0] == 0:
            continue

        if style == "line":
            _plot_bed_bar_l(
                axs[ix],
                bed2plot,
                row["fetch_start"],
                row["fetch_end"],
                needReverse=row["isReverse"],
                style="line",
                pri_col=primary_col[ix],
                #5sec_col=secondary_col[ix],
                alpha=alpha,
            )

        if style == "bar":
            _plot_bed_bar_l(
                axs[ix],
                bed2plot,
                row["fetch_start"],
                row["fetch_end"],
                needReverse=row["isReverse"],
                style="bar",
                pri_col=primary_col[ix],
                sec_col=secondary_col[ix],
                alpha=alpha,
            )

        if style == "rec":
            _plot_bed_rec(
                ax,
                axs[ix],
                bed2plot,
                row["fetch_start"],
                row["fetch_end"],
                needReverse=row["isReverse"],
                color=primary_col[ix],
                cname=cmap[ix],
                alpha=alpha,
                vmin=vmin,
                vmax=vmax,
                score_label=score_label,
                intervals=intervals,
                score_label_size=score_label_size,
            )

        if style == "tri":
            _plot_bed_tri(
                ax,
                axs[ix],
                bed2plot,
                row["fetch_start"],
                row["fetch_end"],
                needReverse=row["isReverse"],
                color=primary_col[ix],
                cname=cmap[ix],
                alpha=alpha,
                vmin=vmin,
                vmax=vmax,
                score_label=score_label,
                score_label_size=score_label_size,
            )

            if invert_y:
                axs[ix].set_ylim(max_len / 2, 0)
            else:
                axs[ix].set_ylim(0, max_len / 2)

        if style == "link":
            _plot_bed_link(
                ax=ax,
                bed=bed2plot,
                start=row["fetch_start"],
                end=row["fetch_end"],
                needReverse=row["isReverse"],
                invert_y=invert_y,
                color=primary_col[ix],
                cmap=cmap[ix],
                alpha=alpha,
            )

    if style in ["line", "bar"]:
        for axi in axs:
            if invert_y:
                axi.set_ylim(vmax, vmin)
            else:
                axi.set_ylim(vmin, vmax)

        if invert_y:
            ax.set_ylim(vmax, vmin)
        else:
            ax.set_ylim(vmin, vmax)

        text_label_y_pos = vmax
        if invert_y:
            text_label_y_pos = vmin
            ax.text(
                0,
                vmax,
                " [{1}, {0}]".format(tick_fl % vmin, tick_fl % vmax),
                va="bottom",
                fontsize=tick_fontsize,
            )
        else:
            ax.text(
                0,
                text_label_y_pos,
                " [{0}, {1}]".format(tick_fl % vmin, tick_fl % vmax),
                va="top",
                fontsize=tick_fontsize,
            )


def _make_tri_data(start, end):
    data = np.array(
        [[start, 0], [end, 0], [start + (end - start) / 2, (end - start) / 2]]
    )
    return data


def _plot_bed_tri(
    mainAX,
    ax,
    bed,
    start,
    end,
    needReverse,
    color,
    cname,
    alpha,
    vmin,
    vmax,
    score_label=None,
    score_label_size=8,
):
    colors = color
    norm = None
    if cname != None:
        if isinstance(cname, str):
            map_vir = cm.get_cmap(cname)
        if isinstance(cname, Colormap):
            map_vir = cname

        # 因为 y 大到一定程度超过临界数值后颜色就会饱和不变(不使用循环colormap)。
        norm = plt.Normalize(vmin, vmax)
        if vmin<0 and vmax>0:
            norm = TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        
        # matplotlib.colors.Normalize 对象，可以作为参数传入到绘图方法里
        # 也可给其传入数值直接计算归一化的结果
        norm_y = norm(bed["score"])
        colors = map_vir(norm_y)

    patches = []
    bed = bed.reset_index()
    # print(colors[0])
    # print(colors[1])

    for i, row in bed.iterrows():
        polygon = Polygon(
            _make_tri_data(row["start"], row["end"]), closed=True, color=colors
        )
        patches.append(polygon)

    p = PatchCollection(patches, alpha=alpha, match_original=True)
    ax.add_collection(p)

    xlim_s = start
    xlim_e = end
    if needReverse == True:
        xlim_s = end
        xlim_e = start

    ax.set_xlim(xlim_s, xlim_e)
    if cname != None:
        sm = cm.ScalarMappable(norm=norm, cmap=map_vir)
        cax = mainAX.inset_axes([1.01, 0, 0.01, 1])
        cb = plt.colorbar(sm, ax=mainAX, cax=cax, label=score_label)
        cb.set_label(score_label, fontsize=score_label_size)


def _plot_bed_rec(
    mainAX,
    ax,
    bed,
    start,
    end,
    needReverse,
    color,
    cname,
    alpha,
    vmin,
    vmax,
    score_label=None,
    intervals=1,
    score_label_size=8,
):
    colors = color
    if cname != None:
        if isinstance(cname, str):
            map_vir = cm.get_cmap(cname)
        if isinstance(cname, Colormap):
            map_vir = cname

        # 因为 y 大到一定程度超过临界数值后颜色就会饱和不变(不使用循环colormap)。
        norm = plt.Normalize(vmin, vmax)
        if vmin<0 and vmax>0:
            norm = TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        
        # matplotlib.colors.Normalize 对象，可以作为参数传入到绘图方法里
        # 也可给其传入数值直接计算归一化的结果
        norm_y = norm(bed["score"])
        colors = map_vir(norm_y)
    bed.loc[:, "length"] = abs(bed["end"] - bed["start"])
    bottom = list(range(intervals))
    # broadcast
    if intervals < bed.shape[0]:
        repeat_times = (bed.shape[0] + intervals - 1) // intervals
        bottom = (bottom * repeat_times)[: bed.shape[0]]
    else:
        bottom = bottom[: bed.shape[0]]
    bed.loc[:, "bottom"] = bottom

    ax.bar(
        x="start",
        width="length",
        height=1,
        bottom="bottom",
        align="edge",
        color=colors,
        edgecolor=None,
        data=bed,
        alpha=alpha,
    )
    xlim_s = start
    xlim_e = end
    if needReverse == True:
        xlim_s = end
        xlim_e = start
    ax.set_xlim(xlim_s, xlim_e)
    if cname != None:
        #sm = cm.ScalarMappable(norm=norm, cmap=map_vir)
        sm = cm.ScalarMappable(norm=norm, cmap=map_vir)
        cax = mainAX.inset_axes([1.01, 0, 0.01, 0.9])
        cb = plt.colorbar(sm, ax=mainAX, cax=cax, label=score_label)
        cb.set_ticks([vmin, vmax])
        cb.set_label(score_label, fontsize=score_label_size)


def _plot_bed_bar_l(
    ax, bed, start, end, needReverse, style="bar", pri_col="#3271B2", sec_col="#FBD23C", alpha=1
):
    plot_bottom_line = True
    if style == "bar":
        bed_pos = bed.query('score>=0')
        bed_neg = bed.query('score<0')
        if bed_neg.shape[0] >0 :
            plot_bottom_line=False
        ax.bar(
            x=bed_pos["start"],
            width=bed_pos["end"] - bed_pos["start"],
            height=bed_pos["score"],
            color=pri_col,
            alpha=alpha,
            align="edge",
        )
        ax.bar(
            x=bed_neg["start"],
            width=bed_neg["end"] - bed_neg["start"],
            height=bed_neg["score"],
            color=sec_col,
            alpha=alpha,
            align="edge",
        )

    if style == "line":
        ax.plot(
            bed["start"], bed["score"], color=pri_col, alpha=alpha, solid_capstyle="butt"
        )

    xlim_s = start
    xlim_e = end
    if needReverse == True:
        xlim_s = end
        xlim_e = start
    ax.set_xlim(xlim_s, xlim_e)

    spines = ["top", "bottom", "left", "right"]
    if needReverse == True:
        if plot_bottom_line==True:
            del spines[0]
    else:
        if plot_bottom_line==True:
            del spines[1]

    """
    for i in ["top", "right", "left"]:
        ax.spines[i].set_color("none")
        ax.spines[i].set_linewidth(0)

    if plot_bottom_line:
        ax.spines["bottom"].set_color("black")
        ax.spines["bottom"].set_linewidth(1)
    else:
        ax.spines['bottom'].set_color("none")
        ax.spines['bottom'].set_linewidth(0)
    """
    # ax.tick_params(bottom =True,top=False,left=False,right=False)
    # ax.set_xticklabels("")
    # ax.set_yticklabels("")


def _plot_bed_link(
    ax,
    bed,
    start,
    end,
    needReverse=False,
    invert_y=False,
    color="tab:blue",
    cmap="RdBu",
    alpha=1,
):
    bed = bed[bed.columns[[0, 1, 2]]]
    bed.columns = ["chrom", "start", "end"]
    bed["length"] = bed["end"] - bed["start"]
    max_extend = 0
    if bed.shape[0] > 0:
        if max_extend < max(bed["length"]):
            max_extend = max(bed["length"])

    _plot_loop_arc(ax, bed, color, max_extend, invert_y, start, end, "start", "end")

    if needReverse == True:
        ax.set_xlim(end, start)
    else:
        ax.set_xlim(start, end)

    spines = ["top", "bottom", "left", "right"]
    if invert_y == True:
        del spines[0]
    else:
        del spines[1]
    for i in spines:
        ax.spines[i].set_visible(False)

    ax.set_xticks([])
    ax.set_xticklabels("")
    ax.set_yticks([])
    ax.set_yticklabels("")



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.cm import get_cmap
from matplotlib.colors import LinearSegmentedColormap, ListedColormap


def hex2rgb(value):
    # convert hex to rgb
    value = value.lstrip("#")
    lv = len(value)
    rgb = tuple(int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3))
    return np.array([rgb[0] / 255, rgb[1] / 255, rgb[2] / 255, 1])


def colorC(
    cname="RdBu_r",
    bottom_color="#ffffff",
    bad_color="white",
    over_color="white",
    under_color="white",
    alpha=0,
):
    # cname = ["twilight_shifted", "jet", "RdBu_r", "RdGy_r", "BrBG_r", "hot_r", "Spectral_r"]
    # cname = ["terrain_r", "ocean", "gist_earth", "gist_stern_r", "tab20b", "twilight"]
    from matplotlib.colors import ListedColormap

    if isinstance(cname, str):
        # cmap=plt.get_cmap(cname)
        cmap = get_cmap(cname)
    else:
        cmap = cname

    cmap.set_bad(color=bad_color, alpha=alpha)
    cmap.set_over(color=over_color, alpha=alpha)
    cmap.set_under(color=under_color, alpha=alpha)

    if bottom_color == None:
        return cmap

    bottom_color = hex2rgb(bottom_color)
    newcolors = cmap(np.linspace(0, 1, 256))
    # white = np.array([1, 1, 1, 1])
    newcolors[0, :] = bottom_color
    newcmap = ListedColormap(newcolors)

    return newcmap


fruitpunch = sns.blend_palette(["white", "red"], as_cmap=True)
# fruitpunch2 = sns.blend_palette(['white', 'blue'], as_cmap=True)
fruitpunch2 = sns.blend_palette(["white", "purple"], as_cmap=True)
washu = sns.blend_palette(["white", "#B9378A"], as_cmap=True)
fruitpunch3 = LinearSegmentedColormap.from_list(
    "fruitpunch3", [(0, "white"), (0.03, "w"), (0.4, "r"), (1, "#CF3F35")], N=100
)


# -----------------------------------------------------------------



import matplotlib.pyplot as plt

plt.rcParams["pdf.fonttype"] = 42
plt.rcParams["svg.fonttype"] = "none"

from typing import List, Union

from matplotlib.axes import Axes


def make_spec(
    figsize: tuple = (8, 3),
    height_ratios: Union[List[float], None] = None,
    width_ratios: Union[List[float], None] = None,
    hspace: float = 0,
    wspace: float = 0,
):
    """
    Make `GridSpec` allows complex layout of Axes in the figure.

    Parameters
    ----------
    figsize : 2-tuple of floats, default: `figure.figsize`
            Figure dimension ``(width, height)`` in inches.
            ``Whole Figure Size``
    width_ratios : `array-like`
        if `height_ratios` is setting, this parameter will ignore.
        Defines the relative widths of the columns. Each column gets a
        relative width of ``width_ratios[i] / sum(width_ratios)``.
        If not given, all columns will have the same width.
    height_ratios : `array-like`
        Defines the relative heights of the rows. Each row gets a
        relative height of ``height_ratios[i] / sum(height_ratios)``.
        If not given, all rows will have the same height.
    wspace, hspace : `float`, default: 0
        The amount of width/height reserved for space between subfigures,
        expressed as a fraction of the average subfigure width/height.
        If not given, the values will be inferred from a figure or
        rcParams when necessary.
    Returns
    -------
    Figure, list[Axes]
        `Figure`: top level `~matplotlib.artist.Artist`, which holds all plot elements.
            Many methods are implemented in `FigureBase`.
        `list[Axes]`: is a list of the Axes objects.  The order of the axes is left-to-right
            or top-to-bottom of their position in the total layout.

    Example
    -------
    >>> import trackc as tc
    >>> fig, axs = tc.make_spec(figsize=(6,6), height_ratios=[1,1,1])
    >>> axs[0].plot([1,2,3])
    >>> tc.savefig('test.pdf')
    """

    nfig = 0
    nrows = 1
    ncols = 1
    ratios = height_ratios
    if height_ratios != None and isinstance(height_ratios, list):
        nfig = len(height_ratios)
        width_ratios = None
        nrows = nfig

    if width_ratios != None and isinstance(width_ratios, list):
        nfig = len(width_ratios)
        height_ratios = None
        ncols = nfig
        ratios = width_ratios

    if nfig <= 0:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        return fig, ax

    fig = plt.figure(figsize=figsize)
    spec = fig.add_gridspec(
        nrows=nrows,
        ncols=ncols,
        left=None,
        bottom=None,
        right=None,
        top=None,
        wspace=wspace,
        hspace=hspace,
        width_ratios=width_ratios,
        height_ratios=height_ratios,
    )

    axs = [None] * nfig
    for i, v in enumerate(ratios):
        axs[i] = fig.add_subplot(spec[i])

    return fig, axs


# fig, axs = make_spec(height_ratios=[1,1.5,2.4], hspace=0.1)


class mortise:
    bottom = 0
    top = 0
    track_height = 0
    hspace = 0
    ax = None

    def __init__(self, ax_ref: Axes, bottom: float, track_height: float, hspace: float):
        self.ax = ax_ref.inset_axes([0, bottom, 1, track_height])
        self.bottom = bottom
        self.top = bottom + track_height
        self.track_height = track_height
        self.hspace = hspace


class tenon:
    """
    Make a virtual-figure, based on virtual-figure, users can `add` `Axes` to
    the `top or bottom` of the total layout. After adding `Axes`, the user can use the ``axs`` method
    to select an axes from the added axes list. The order of the axes is top-to-bottom of
    the position in the total layout.

    Just like Lego blocks, you can add new blocks above (top) or below (bottom)

    Example
    -------
    >>> import trackc as tc
    >>> ten = trackc.tenon(figsize=(6,1))
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='top', height=1, hspace=0.1)
    >>> ten.axs(0).plot([1,2,3])
    >>> ten.axs(1).plot([1,1])
    >>> tc.savefig('test.pdf')
    """

    mortises = []
    fig = None
    ax = None

    def __init__(self, figsize=(7, 1)):
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        ax.set_axis_off()
        self.fig = fig
        self.ax = ax

    def show(self):
        for i, v in enumerate(self.mortises):
            # v.ax.set_ylabel('mortises[{0}]'.format(i), rotation=0)
            v.ax.text(0.5, 0.5, ".axs({0})".format(i), ha="center", va="center")
            v.ax.set_xticklabels([])
            v.ax.set_xticks([])
        self.fig.show()

    def add(self, pos="top", height=1, hspace=0):
        """
        Add a `Axes` to the total layout.
        Just like Lego blocks, you can add new blocks above (top) or below (bottom)

        Parameters
        ----------
        pos
            one option of ['top', 'bottom']
            Choose whether to add a new Axes at the top or bottom.
            If set top, then a `~matplotlib.axes.Axes` object will add to top of the top `~matplotlib.axes.Axes`
        height: `float`
            the relative height of the newly added `Axes` compared to the virtual-figure.
            For example, if the virtual-figure has a height of 1 and the new Axes has a height of 4,
            the actual height of the subplot will be 1 * 4.
        hspace: `float`
            The amount of height reserved for space between the top or bottom subfigures
        """
        if len(self.mortises) != 0:
            ix = -1
            if pos == "top":
                ix = 0
            ref_lego_ax = self.mortises[ix]
        else:
            ref_lego_ax = mortise(self.ax, 0, height, 0)
            self.mortises = [ref_lego_ax]
            return

        bottom = ref_lego_ax.bottom - hspace - height
        if pos == "top":
            bottom = ref_lego_ax.top + hspace

        lego_ax_demo = mortise(self.ax, bottom, height, hspace)
        if pos == "top":
            self.mortises.insert(0, lego_ax_demo)
        else:
            self.mortises.append(lego_ax_demo)

    def _remove(self, pos="top"):
        """
        Remove a `Axes` to the total layout.
        Just like Lego blocks, you can remove new blocks from above (top) or below (bottom)

        Parameters
        ----------
        pos
            Option of ['top', 'bottom']
            Choose whether to remove a Axes from top or bottom of the whole layout.
        """
        if len(self.mortises) >= 1:
            if pos == "top":
                self.mortises.remove(self.mortises[0])
            elif pos == "bottom":
                self.mortises.remove(self.mortises[-1])
            else:
                pass

    def axs(self, index):
        """
        Get one Axes from the total layout.

        Parameters
        ----------
        index
            The Axes list element index

        Returns
        -------
        `~matplotlib.axes.Axes` object
        """
        return self.mortises[index].ax


def savefig(outfile):
    """
    Save the figure.
    Parameters
    ----------
    outfile
        out file path
    """
    # plt.tight_layout()
    plt.savefig(outfile, bbox_inches="tight")
    # plt.savefig(outfile)


def show():
    """
    show the Axes of the layout, help users resize each axes.
    """
    plt.tight_layout()



import sys
from matplotlib.axes import Axes
from typing import Union, Optional, Sequence
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.colors import Colormap
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.colors import LogNorm #,CenteredNorm, SymLogNorm,PowerNorm, Normalize  
from trackc.tl._getRegionsCmat import GenomeRegion
from trackc.pa import fruitpunch, trackcl_9, trackcl_11, zeileis_28

def bdgmat_track(ax: Optional[Axes] = None,
                 bed: Union[pd.DataFrame, str, None]  = None,
            regions: Union[Sequence[str], str, None] = None,
            style: Union[str, None] = 'heatmap',
            logdata: bool = False, 
            minrange: Optional[float] = None,
            maxrange: Optional[float] = None,
            color: Union[Sequence[str], None] = 'silver',
            cmap: Union[Colormap, str, None] = fruitpunch,
            alpha: Union[float, None] = 1,
            label: Union[str, None] = None,
            label_fontsize: Union[int, None] = 8,
            ticks_on: bool=True,
            tick_fontsize: Optional[int] = 7,
            tick_fl: Optional[str] = "%0.2f",
            line_lg_ncol: Optional[int] = 1
            ):
    """\
    Plot bedGraph matrix track, support for multiple or reverse genome regions.
    bedgraph matrix file like bedgraph, columns from 4th are values
    bedgraph matrix should be sorted by chrom, start

    Parameters
    
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    bed: `pd.DataFrame` | `str`
        If ``bed`` if a filepath, the file should have headers, The first three columns are [chrom start end]
        bedGraph matrix format: [chrom start end scorename1 scorename2 ...]
    regions: `str` | `str list`
        The genome regions to plot
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``), 
            which means you want to get the reverse region
    style: `str`
        heatmap or line, default heatmap
    logdata: `bool` | `bool list`
        whether log the data before plotting
    minrange: `float` | `float list`
        the minimum range of values used to show
    maxrange: `float` | `float list` 
        the maximum range of values used to show
    color: `str` or `list`
        the color of the plot for style: line, if color is a list, the line color will be set by matrix columns order
    cmap: `str` | `matplotlib.colors.Colormap`
        the colormap of the plot for style: heatmap
    show_h
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    #axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(bed, str)==True:
        bed=pd.read_table(bed, sep="\t", header=0)

    if style=="line":
        if isinstance(color, list)==False:
            color = [color]
        score_columns = bed.shape[1]-3
        if len(color) < score_columns:
            #repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
            #color = (color * repeat_times)[:score_columns]
            print('color sizes less than plots, use defaults')
            if score_columns <= 9:
                color = trackcl_9
            elif score_columns <= 11:
                color = trackcl_11
            elif score_columns <= 28:
                color = zeileis_28
            else:
                pass


    bed['chrom'] = bed['chrom'].astype(str)
    
    df2plot = pd.DataFrame()
    bed = bed.fillna(0)
    chromos = bed['chrom'].unique()
    for ix, row in line_GenomeRegions.iterrows(): 
        raw_chr = row["chrom"]
        if row["chrom"] not in chromos:
            if row["chrom"].startswith('chr'):
                row["chrom"] = row["chrom"].lstrip('chr')
            else:
                row["chrom"] = 'chr' + row["chrom"]
            if row["chrom"] not in chromos:
                print(f'{raw_chr} not in begmat chroms!')
                #sys.exit(0)
                return

        bed2plot = bed[(bed['chrom']==row['chrom']) & (bed['end']>=row['fetch_start']) & (bed['start']<=row['fetch_end'])]
        if row['isReverse']==True:
            bed2plot = bed2plot.iloc[::-1]
        if df2plot.shape[0]==0:
            df2plot = bed2plot
        else:
            df2plot = pd.concat([df2plot, bed2plot], axis=0)
    
    df2plot = df2plot.iloc[:,3:]
    if minrange==None:
        minrange = df2plot.min().min()
    if maxrange==None:
        maxrange = df2plot.max().max()

    norm = None
    if logdata:
        norm=LogNorm()
    clim = (minrange, maxrange)

    
    if style=="heatmap":
        df2plot = df2plot.T
        caxes = ax.pcolormesh(df2plot, cmap=cmap, clim=clim, edgecolor='none', norm=norm, snap=True, linewidth=.001)
        _colorbar(ax, caxes, tick_font_size=tick_fontsize)
        ax.set_xticks([])  # 隐藏刻度
        ax.set_xticklabels([])  # 隐藏刻度标签
        #if
        ax.set_yticks([i+0.5 for i in range(0, df2plot.shape[0])])  # 隐藏刻度
        if ticks_on:
            ax.set_yticklabels(df2plot.index, fontsize=tick_fontsize)  # 隐藏刻度标签
        else:
            ax.set_yticklabels([])
            ax.set_yticks([])
    elif style=="line":
        #print(df2plot)
        #import sys
        #sys.exit(1)
        df2plot = df2plot.reset_index()
        del df2plot['index']
        line = ax.plot(df2plot)

        if len(color) < df2plot.shape[1]:
            repeat_times = (df2plot.shape[1] + len(color) - 1) // len(color)
            color = (color * repeat_times)

        for i,v in enumerate(line):
            v.set_label(df2plot.columns[i])
            v.set(color=color[i])

        ax.set_xlim(0, df2plot.shape[0]-1)
        print(f'bdg_mat minrange:{minrange} maxrange:{maxrange}')
        ax.set_xticklabels([])
        ax.set_xticks([])
        ax.set_ylim([minrange, maxrange])
        ax.set_yticks([minrange, maxrange])
        ax.set_yticklabels([f'{tick_fl % minrange}', f'{tick_fl % maxrange}'], fontsize=tick_fontsize)  # 隐藏刻度标签
        if ticks_on:
            ax.legend(loc='best', bbox_to_anchor=(1, 0, 0, 1), fontsize=tick_fontsize, ncol=line_lg_ncol)

    ax.set_ylabel(label, fontsize=label_fontsize)

def _colorbar(axm,im, width="3%", height="100%", tick_font_size=7):
    axins = inset_axes(axm,
                   width=width,  # width = 5% of parent_bbox width
                   height=height,  # height : 50%
                   loc='lower left',
                   bbox_to_anchor=(1.01, 0, 1, 1),
                   bbox_transform=axm.transAxes,
                   borderpad=0,
                   )
    cbar=plt.colorbar(im, cax=axins, orientation='vertical')
    cbar.set_ticks([im.get_array().min(), im.get_array().max()])
    cbar.ax.tick_params(labelsize=tick_font_size)
    cbar.ax.yaxis.tick_right()
    cbar.ax.spines['top'].set_color('none')
    cbar.ax.spines['right'].set_color('none')
    cbar.ax.spines['bottom'].set_color('none')
    cbar.ax.spines['left'].set_color('none')




from typing import Optional, Sequence, Union

import pandas as pd
from matplotlib.axes import Axes

from trackc.pa import trackcl_11
from trackc.tl._getRegionsCmat import GenomeRegion


def zoomin(
    ax: Optional[Axes] = None,
    raw_regions: Union[Sequence[str], str, None] = None,
    zoomin_regions: Union[Sequence[str], str, None] = None,
    colors: Union[Sequence[str], None] = None,
    alpha: float = 1,
    line_on: bool = False,
    fill: bool = True,
):
    """
    Plot zoomin track, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    raw_regions: `str` | `str list`
        The raw genome regions, some of these regions will be selected to zoom in.
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``

        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``), which means the reverse region
    zoomin_regions: `str` | `str list`
        regions to be zoomin, The format is the same as `raw_regions`.
        the regions of `zoomin_regions` should be located in the `raw_regions`
    colors: `str list`
        the colors of the zoomin_regions
    alpha: `float`
    line_on: `bool`
        Whether to display the line of zoomin plots
    fill: `bool`
        Whether to fill the zoomin plots

    Example
    -------
    >>> full_regions = "chr18:45000000-78077248"
    >>> zoom_regions = ['chr18:47400000-48280000', 'chr18:75280000-74030000']
    >>> neo_domain_regions = ['chr18:47950000-48280000', 'chr18:75280000-74850000']

    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='bottom', height=1, hspace=0.1)

    >>> tc.pl.zoomin(ax=ten.axs(0), raw_regions=full_regions, zoomin_regions=zoom_regions, line_on=True, fill=False, alpha=0.5)
    >>> tc.pl.zoomin(ax=ten.axs(1), raw_regions=zoom_regions, zoomin_regions=neo_domain_regions, line_on=False, fill=True, alpha=0.5)
    >>> tc.savefig('trackc_zoomin_track.pdf')
    """

    if isinstance(raw_regions, str):
        raw_regions = [raw_regions]
    if isinstance(zoomin_regions, str):
        zoomin_regions = [zoomin_regions]

    row_GRs = _region_pos(raw_regions)
    zoomin_GRs = _region_pos(zoomin_regions)
    zoomin_GRs["ors"] = None
    zoomin_GRs["ore"] = None
    zoomin_GRs_cp = zoomin_GRs.copy()
    for i, row in zoomin_GRs_cp.iterrows():
        start, end = _get_zoom_origin_pos(
            row["chrom"], row["fetch_start"], row["fetch_end"], row_GRs
        )
        ss = start
        ee = end
        if row["isReverse"] == True:
            ss = end
            ee = start
        zoomin_GRs.loc[i, "ors"] = ss
        zoomin_GRs.loc[i, "ore"] = ee

    zoomin_GRs = zoomin_GRs.reset_index()

    if colors == None:
        colors = trackcl_11

    for i, row in zoomin_GRs.iterrows():
        x = row[["ps", "pe", "ore", "ors"]]
        y = [0, 0, 1, 1]

        if fill == True:
            ax.fill(x, y, color=colors[i], alpha=alpha)

        if line_on == True:
            ax.plot(
                (x[0], x[3]),
                (y[0], y[3]),
                color=colors[i],
                alpha=alpha,
                solid_capstyle="butt",
            )
            ax.plot(
                (x[1], x[2]),
                (y[1], y[2]),
                color=colors[i],
                alpha=alpha,
                solid_capstyle="butt",
            )

    ax.set_axis_off()
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])


def _region_pos(regions):
    GRs = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    GRs["len"] = GRs["fetch_end"] - GRs["fetch_start"]
    GRs["pos_e"] = GRs["len"].cumsum()
    GRs["pos_s"] = GRs["pos_e"] - GRs["len"]
    GRs_fulllen = GRs["len"].sum()
    GRs["ps"] = GRs["pos_s"] / GRs_fulllen
    GRs["pe"] = GRs["pos_e"] / GRs_fulllen
    return GRs


def _get_zoom_origin_pos(zoom_chrom, zooms, zoome, origin_pos_df):
    fulllen = origin_pos_df["len"].sum()
    origin_pos = origin_pos_df[origin_pos_df["chrom"] == zoom_chrom]
    origin_pos = origin_pos[
        (zooms >= origin_pos["fetch_start"]) & (zoome <= origin_pos["fetch_end"])
    ]

    start = None
    end = None
    if origin_pos.shape[0] == 0:
        print("{0}:{1}-{2} not in origion regions".format(zoom_chrom, zooms, zoome))
    else:
        ix = origin_pos.index[0]
        start_dis = origin_pos_df.loc[ix, "pos_s"] + abs(
            zooms - origin_pos_df.loc[ix, "start"]
        )
        end_dis = origin_pos_df.loc[ix, "pos_s"] + abs(
            zoome - origin_pos_df.loc[ix, "start"]
        )
        start = start_dis / fulllen
        end = end_dis / fulllen
    return start, end
